import sqlparse

from watchmen_storage.sql_analysis.ast_visitor import CaseVisitor, ComparisonVisitor, FunctionVisitor, \
	IdentifierListVisitor, IdentifierVisitor, ParenthesisVisitor, QueryPerformance, SqlColumn, SqlContext, TokenVisitor, \
	WhereVisitor


class SqlParser:
	def __init__(self):
		self.visitors = [
			IdentifierListVisitor(), FunctionVisitor(), IdentifierVisitor(), ParenthesisVisitor(),
			ComparisonVisitor(), WhereVisitor(), CaseVisitor(), TokenVisitor()
		]

	def find_visitor(self, ast):
		for v in self.visitors:
			if v.support(ast):
				return v

	def start(self):
		pass

	def parse(self, sql: str) -> QueryPerformance:
		context = SqlContext()
		statements = sqlparse.split(sql)
		for statement in statements:
			tokens = sqlparse.parse(statement)[0].tokens
			for index, token in enumerate(tokens):
				visitor = self.find_visitor(token)
				if visitor:
					previous_one = self.get_previous(index, tokens)
					next_one = self.get_next(index, tokens)
					visitor.process(token, context, self.visitors, previous_one, next_one)

		qp = self.process_context_result(context)
		qp.sql = sql
		return qp

	# noinspection PyMethodMayBeStatic
	def remove_duplicate(self, current_column_list):
		result = []
		last_name = None
		for column in current_column_list:
			if last_name != column.name:
				result.append(column)
				last_name = column.name
		return result

	def process_context_result(self, context: SqlContext):
		query_performance = QueryPerformance()
		column_dimensions = self.build_column_dimension(context)
		query_performance.column_dimensions = column_dimensions
		query_performance.topic_dimensions = self.build_table_dimension(context)
		return query_performance

	# noinspection PyMethodMayBeStatic
	def build_table_dimension(self, context: SqlContext):
		tables = context.table
		tables_value = ""
		for table in tables:
			if tables_value == "":
				tables_value = table
			else:
				tables_value = tables_value + "_" + table
		return tables_value

	def build_column_dimension(self, context: SqlContext):
		column_dict = context.columns
		level = context.level
		current_column_list = column_dict["level_" + str(level)]
		dimension_value = ""
		sort_current_column_list = self.remove_duplicate(
			sorted(current_column_list, key=lambda a_column: a_column.name))
		for column in sort_current_column_list:
			column: SqlColumn = column
			if dimension_value == "":
				dimension_value = column.name
			else:
				dimension_value = dimension_value + "_" + column.name
		return dimension_value

	# noinspection PyMethodMayBeStatic
	def get_previous(self, index, tokens):
		if index > 0:
			return tokens[index - 1]
		else:
			return None

	# noinspection PyMethodMayBeStatic
	def get_next(self, index, tokens):
		if index < len(tokens) - 1:
			return tokens[index + 1]
		else:
			return None



from abc import ABC, abstractmethod
from typing import Dict
import logging


logger = logging.getLogger(__name__)


class SecretsManger(ABC):
	
	@abstractmethod
	def ask_secrets(self, secret_id) -> Dict:
		pass



from typing import Union

from watchmen_model.admin import Topic, TopicKind
from .storage_types import ColumnNameLiteral


def as_table_name(topic_or_name_or_column_name_literal: Union[Topic, str, ColumnNameLiteral]) -> str:
	x = topic_or_name_or_column_name_literal
	if isinstance(x, str):
		return f'topic_{x.strip().lower()}'
	elif isinstance(x, ColumnNameLiteral):
		if x.synonym:
			return x.entityName.strip().lower()
		else:
			return f'topic_{x.entityName.strip().lower()}'
	elif x.kind == TopicKind.SYNONYM:
		# use exactly the given name
		return x.name.strip().lower()
	else:
		return f'topic_{x.name.strip().lower()}'



from pydantic import BaseModel
from sqlparse.sql import Case, Comparison, Function, Identifier, IdentifierList, Parenthesis, Token, Where


class SqlContext(BaseModel):
	status: str = None
	sub_status: str = None
	table: list = []
	group_by: list = []
	joins: list = []
	where: list = []
	columns: dict = {"level_0": [], "level_1": [], "level_2": []}
	function: list = []
	level: int = -1


class SqlColumn(BaseModel):
	"""
	Represents a column in a sql select statement
	"""
	name: str = None
	alias: str = None
	function: list = None
	table: str = None


class JoinColumn(BaseModel):
	"""
	Represents a column in a sql join statement
	"""
	left: str = None
	leftTable: str = None
	operator: str = None
	right: str = None
	rightTable: str = None


class JoinTable(BaseModel):
	name: str = None
	alias: str = None
	join_column: str = None


class GroupByColumn(BaseModel):
	name: str = None
	alias: str = None
	table: str = None


class WhereColumn(BaseModel):
	name: str = None
	alias: str = None
	table: str = None
	operator: str = None
	value: str = None


class QueryPerformance(BaseModel):
	topic_dimensions: str = None
	column_dimensions: str = None
	execution_time: int = None
	data_volume: int = None
	join_dimensions: str = None
	where_dimensions: str = None
	group_by_dimensions: str = None
	sql: str = None


def build_function(name):
	if value_is_function(name):
		return name
	else:
		return None


def find_visitor(visitor_list, ast):
	for v in visitor_list:
		if v.support(ast):
			return v


def value_is_function(name):
	return name == 'count' or name == 'sum' or name == 'avg' or name == 'max' or name == 'min' or name == 'distinct'


def add_column(context, value: str):
	column_dict = context.columns
	level = context.level
	key = "level_" + str(level)
	if context.sub_status == "column_topic":
		column = column_dict[key][-1]
		column.name = value
	else:
		column = SqlColumn(name=value)
		column_dict[key].append(column)


def add_column_as(context: SqlContext, value):
	if context.level == 0:
		column_dict = context.columns
		column = column_dict["level_0"][-1]
		column.alias = value
	elif context.level == 1:
		column_dict = context.columns
		column = column_dict["level_1"][-1]
		column.alias = value
	elif context.level == 2:
		column_dict = context.columns
		column = column_dict["level_2"][-1]
		column.alias = value


def find_operator(token: Token):
	"""
	:param token:
	:return:
	"""
	for sub_token in token.tokens:
		if isinstance(sub_token, Token):
			if sub_token.normalized == "=" \
					or sub_token.normalized == ">" \
					or sub_token.normalized == "<" \
					or sub_token.normalized == ">=" \
					or sub_token.normalized == "<=" \
					or sub_token.normalized == "!=":
				return sub_token.normalized


def add_column_topic(context, value):
	"""
	:param context:
	:param value:
	"""
	context.sub_status = "column_topic"
	if context.level == 0:
		column_dict = context.columns
		column = SqlColumn(table=value)
		column_dict["level_0"].append(column)
	elif context.level == 1:
		column_dict = context.columns
		column = SqlColumn(table=value)
		column_dict["level_1"].append(column)
	elif context.level == 2:
		column_dict = context.columns
		column = SqlColumn(table=value)
		column_dict["level_2"].append(column)


def set_status(context: SqlContext, token: Token):
	if token.is_keyword and token.value.upper() == 'SELECT':
		context.status = "select"
		context.level = context.level + 1
	elif token.is_keyword and token.value.upper() == 'FROM':
		context.status = "from"
	elif token.is_keyword and token.value.upper() == "AS":
		if context.status == "from":
			context.status = "table_as"
		else:
			context.status = "as"
	elif token.is_keyword and token.value.upper() == "GROUP BY":
		context.status = "group_by"
	elif token.is_keyword and token.value.upper() == "INNER JOIN":
		context.status = "join"
	elif token.is_keyword and token.value.upper() == "ON":
		context.status = "on"
	# elif token.value == "IN" or token.value == "NOT IN" or token.value == "EXISTS" or token.value == "NOT EXISTS":
	elif context.status == "case":
		context.sub_status = token.value
	elif token.is_keyword or value_is_function(token.value):
		function = build_function(token.value)
		context.function.append(function)


def get_current_column(context):
	column_dict = context.columns
	level = context.level
	key = "level_" + str(level)
	return column_dict[key][-1]


class TokenVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Token)

	# noinspection PyUnusedLocal
	def process(self, token: Token, context, visitor_list, previous_token: Token = None, next_token: Token = None):
		if self.ignore_value(token):
			return None
		elif token.value == ")":
			if context.sub_status == "start_function":
				self.process_end_function(context)
			else:
				return None
		elif token.is_keyword or value_is_function(token.value):
			set_status(context, token)
		elif not token.is_keyword \
				and not token.is_group \
				and not token.is_whitespace \
				and not token.value.startswith("anon"):
			if token.value == ".":
				context.status = "dot"
			elif (
					context.sub_status == "IN"
					or context.sub_status == "THEN"
					or context.sub_status == "ELSE"
					or context.sub_status == "END") \
					and context.status == "case":
				return None
			elif token.value == "IN" and context.status == "case":
				context.sub_status = "IN"
			elif context.status == "from" or context.status == "join":
				context.table.append(token.value)
			elif context.status == "group_by":
				context.group_by.append(token.value)
			elif context.status == "as":
				add_column_as(context, token.value)
				context.status = None
			elif context.status == "case" and context.sub_status == "when":
				if token.value.startswith("topic_"):
					add_column_topic(context, token.value)
				else:
					add_column(context, token.value)
			else:
				if token.value.startswith("topic_") and not context.status == "where":
					add_column_topic(context, token.value)
				else:
					add_column(context, token.value)

	# noinspection PyMethodMayBeStatic
	def process_end_function(self, context):
		context.sub_status = "end_function"
		function_list = context.function
		column = get_current_column(context)
		column.function = function_list
		context.function = []

	# noinspection PyMethodMayBeStatic
	def ignore_value(self, token):
		return token.is_whitespace \
			or token.value == "(" \
			or token.value == "." \
			or token.value == "," \
			or token.value == ";" \
			or token.value == "##"


class IdentifierVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Identifier)

	# noinspection PyMethodMayBeStatic
	def process(self, token: Identifier, context, visitor_list, previous_token: Token = None, next_token: Token = None):
		if token.is_group:
			for sub_token in token.tokens:
				visitor = find_visitor(visitor_list, sub_token)
				if visitor:
					visitor.process(sub_token, context, visitor_list, previous_token, next_token)


class ParenthesisVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Parenthesis)

	# noinspection PyMethodMayBeStatic
	def process(
			self, token: Parenthesis, context, visitor_list, previous_token: Token = None,
			next_token: Token = None):
		if context.sub_status == "function":
			context.sub_status = "start_function"
		if (context.sub_status == "IN" or context.sub_status == "THEN") and context.status == "case":
			return None
		if token.M_OPEN and token.M_CLOSE:
			for sub_token in token.tokens:
				visitor = find_visitor(visitor_list, sub_token)
				if visitor:
					visitor.process(sub_token, context, visitor_list, previous_token, next_token)


class FunctionVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Function)

	# noinspection PyMethodMayBeStatic
	def process(self, token: Function, context, visitor_list, previous_token: Token = None, next_token: Token = None):
		# if context.status == "select":
		context.sub_status = "function"
		if token.is_group:
			for sub_token in token.tokens:
				visitor = find_visitor(visitor_list, sub_token)
				if visitor:
					visitor.process(sub_token, context, visitor_list, previous_token, next_token)


class IdentifierListVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, IdentifierList)

	# noinspection PyMethodMayBeStatic
	def process(
			self, token: IdentifierList, context, visitor_list, previous_token: Token = None,
			next_token: Token = None):
		if token.is_group:
			for sub_token in token.tokens:
				visitor = find_visitor(visitor_list, sub_token)
				if visitor:
					visitor.process(sub_token, context, visitor_list, previous_token, next_token)


class WhereVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Where)

	# noinspection PyMethodMayBeStatic
	def process(
			self, token: IdentifierList, context, visitor_list, previous_token: Token = None,
			next_token: Token = None):
		context.status = "where"
		if token.is_group:
			for sub_token in token.tokens:
				visitor = find_visitor(visitor_list, sub_token)
				if visitor:
					visitor.process(sub_token, context, visitor_list, previous_token, next_token)


class ComparisonVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Comparison)

	# noinspection PyMethodMayBeStatic, PyUnusedLocal
	def process(
			self, token: Comparison, context: SqlContext, visitor_list, previous_token: Token = None,
			next_token: Token = None):
		if context.status == "case" and context.sub_status == "WHEN":
			left_table, left = token.left.normalized.split(".")
			add_column_topic(context, left_table)
			add_column(context, left)
		elif context.status == "where":
			if "." in token.left.normalized:
				left_table, left = token.left.normalized.split(".")
				operator = find_operator(token)
				value = token.right.value
				context.where.append(WhereColumn(table=left_table, name=left, operator=operator, value=value))


		elif token.is_group and token.left and token.right and not context.status == "case":
			left_table, left = token.left.normalized.split(".")
			right_table, right = token.right.normalized.split(".")
			operator = find_operator(token)
			context.joins.append(
				JoinColumn(leftTable=left_table, left=left, rightTable=right_table, right=right, operator=operator))


class CaseVisitor:
	# noinspection PyMethodMayBeStatic
	def support(self, ast):
		return isinstance(ast, Case)

	# noinspection PyMethodMayBeStatic
	def process(
			self, token: Case, context, visitor_list, previous_token: Token = None,
			next_token: Token = None):
		context.status = "case"
		if token.is_group:
			for sub_token in token.tokens:
				visitor = find_visitor(visitor_list, sub_token)
				if visitor:
					visitor.process(sub_token, context, visitor_list, previous_token, next_token)




from .storage_spi import TransactionalStorageSPI
from .storage_types import EntityDeleter, EntityCriteriaExpression, ColumnNameLiteral
from .storage_based_worker_id_generator import SNOWFLAKE_WORKER_ID_TABLE, COMPETITIVE_WORKER_SHAPER


class StorageBasedWorkerIdService:

	def __init__(self, storage: TransactionalStorageSPI):
		self.storage = storage

	def release_worker(self, data_center_id: int, worker_id: int) -> None:
		self.storage.begin()
		try:
			self.storage.delete(
				EntityDeleter(
					name=SNOWFLAKE_WORKER_ID_TABLE,
					shaper=COMPETITIVE_WORKER_SHAPER,
					criteria=[
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='data_center_id'), right=data_center_id),
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='worker_id'), right=worker_id)
					]
				)
			)
			self.storage.commit_and_close()
		except Exception as e:
			self.storage.rollback_and_close()
			raise e
		finally:
			self.storage.close()


def get_storage_based_worker_id_service(storage: TransactionalStorageSPI) -> StorageBasedWorkerIdService:
	return StorageBasedWorkerIdService(storage)



from abc import abstractmethod
from enum import Enum
from typing import Optional, List, Dict, Any, Callable

from pydantic import BaseModel

from watchmen_model.system import DataSource, DataSourceParam
from watchmen_utilities import ArrayHelper, is_blank
from .secrets_manager import SecretsManger
from .storage_spi import StorageSPI, TopicDataStorageSPI
from watchmen_utilities import is_not_blank

# database connection information
HOST = "host"
PORT = "port"
USERNAME = "username"
PASSWORD = "password"
NAME = "name"

# datasource params of secret manger
SECRET_TYPE = "secret_type"
AWS_REGION_NAME = "region_name"
AWS_ACCESS_KEY_ID = "access_key_id"
AWS_ACCESS_SECRET_ID = "access_secret_id"
SECRET_ID = "secret_id"

SECRET_MANGER_KEY = [HOST, PORT, USERNAME, PASSWORD, NAME, SECRET_TYPE, AWS_REGION_NAME, AWS_ACCESS_KEY_ID, AWS_ACCESS_SECRET_ID, SECRET_ID]
CONNECTION_INFO_KEY = [HOST, PORT, USERNAME, PASSWORD, NAME]


class SecretType(str, Enum):
	AWS = 'aws'
	ALIYUN = 'aliyun'
	AZURE = 'azure'


class EngineParams(BaseModel):
	host: str
	port: str
	username: str
	password: str
	name: str
	params: List[DataSourceParam] = None


def redress_url(value: str) -> str:
	if value is None:
		return ''
	else:
		return value.strip()


def use_secret(data_source_params: Optional[List[DataSourceParam]]) -> bool:
	return ArrayHelper(data_source_params).find(lambda x: x.name == SECRET_TYPE and x.value is not None) is not None


def ask_secret(data_source_params: Optional[List[DataSourceParam]]) -> Dict:
	return build_secrets_manager(data_source_params).ask_secrets(
		ArrayHelper(data_source_params).find(lambda x: x.name == SECRET_ID).value)


def redress_engine_params(data_source: DataSource) -> EngineParams:
	if use_secret(data_source.params):
		secrets = ask_secret(data_source.params)
		return EngineParams(
			params=ArrayHelper(data_source.params).filter(lambda x: x.name not in SECRET_MANGER_KEY).to_list(),
			**ArrayHelper(data_source.params).filter(lambda x: x.name in CONNECTION_INFO_KEY).to_map(lambda x: x.name,
			                                                                                         lambda x: secrets.get(x.value)))
	else:
		return EngineParams(host=data_source.host,
		                    port=data_source.port,
		                    username=data_source.username,
		                    password=data_source.password,
		                    name=data_source.name,
		                    params=data_source.params)


def ask_datasource_name(data_source: DataSource) -> str:
	return redress_engine_params(data_source).name


class DataSourceHelper:

	def __init__(self, data_source: DataSource):
		self.dataSource = data_source

	@abstractmethod
	def acquire_storage(self) -> StorageSPI:
		pass

	@abstractmethod
	def acquire_topic_data_storage(self) -> TopicDataStorageSPI:
		pass

	@staticmethod
	@abstractmethod
	def acquire_engine_by_url(url: str, params: Any) -> Any:
		pass

	@staticmethod
	@abstractmethod
	def acquire_engine_by_params(username: str, password: str, host: str, port: str, name: str,
	                             data_source_params: Optional[List[DataSourceParam]],
	                             params: Any) -> Any:
		pass

	def ask_acquire_engine_by_url(self) -> Callable[[str, List[Any]], Any]:
		return self.acquire_engine_by_url

	def ask_acquire_engine_by_params(self) -> Callable[[str, str, str, str, str, List[DataSourceParam], Any], Any]:
		return self.acquire_engine_by_params

	def acquire_engine(self, params: Any) -> Any:
		data_source = self.dataSource
		url = redress_url(data_source.url)
		if len(url) != 0:
			return self.ask_acquire_engine_by_url()(url, params)
		else:
			engine_params = redress_engine_params(data_source)
			return self.ask_acquire_engine_by_params()(
				engine_params.username, engine_params.password,
				engine_params.host, engine_params.port,
				engine_params.name,
				engine_params.params,
				params
			)

	# noinspection DuplicatedCode
	@staticmethod
	def find_param(params: Optional[List[DataSourceParam]], key: str) -> Optional[str]:
		if params is None:
			return None

		for param in params:
			if is_not_blank(param.name) and param.name.strip().lower() == key:
				value = param.value
				if is_not_blank(value):
					return value.strip()
		return None


def build_secrets_manager(data_source_params: Optional[List[DataSourceParam]]) -> SecretsManger:
	secret_params = ArrayHelper(data_source_params).filter(lambda x: x.name in SECRET_MANGER_KEY).to_map(lambda x: x.name,
	                                                                                                  lambda x: x.value)
	if secret_params.get(SECRET_TYPE) == SecretType.AWS:
		from .secrets_manager_aws import AmazonSecretsManger
		return AmazonSecretsManger(secret_params.get(AWS_REGION_NAME),
		                           secret_params.get(AWS_ACCESS_KEY_ID),
		                           secret_params.get(AWS_ACCESS_SECRET_ID))
	if secret_params.get(SECRET_TYPE) == SecretType.ALIYUN:
		pass  # todo
	if secret_params.get(SECRET_TYPE) == SecretType.AZURE:
		pass  # todo
	raise NotImplementedError(f'Secret type {secret_params.get(SECRET_TYPE)} is not supported yet.')


def build_aws_secrets_manager(region_name: str,
                              access_key_id: str = None,
                              access_secret_id: str = None) -> SecretsManger:
	if is_blank(region_name):
		raise ValueError(f'region name can not be blank')

	from .secrets_manager_aws import AmazonSecretsManger
	return AmazonSecretsManger(region_name, access_key_id, access_secret_id)


def build_aliyun_secrets_manager(region_name: str,
                                 access_key_id: str = None,
                                 access_secret_id: str = None) -> SecretsManger:
	pass


def build_azure_secrets_manager(region_name: str,
                                access_key_id: str = None,
                                access_secret_id: str = None) -> SecretsManger:
	pass



from typing import Dict
import boto3
import json
from .secrets_manager import SecretsManger
import logging


logger = logging.getLogger(__name__)


class AmazonSecretsManger(SecretsManger):
	def __init__(self, region_name: str, access_key_id: str = None, access_secret_id: str = None):
		self.client = boto3.client(
			service_name='secretsmanager',
			region_name=region_name,
			aws_access_key_id=access_key_id,
			aws_secret_access_key=access_secret_id
		)

	def ask_secrets(self, secret_id) -> Dict:
		try:
			response = self.client.get_secret_value(
				SecretId=secret_id
			)
			return json.loads(response['SecretString'])
		except Exception as e:
			logger.error(e, exc_info=True, stack_info=True)
			raise RuntimeError(e)


from pydantic import BaseSettings
from logging import getLogger

logger = getLogger(__name__)


class StorageSettings(BaseSettings):
	DECIMAL_INTEGRAL_DIGITS: int = 24
	DECIMAL_FRACTION_DIGITS: int = 8
	DISABLE_COMPILED_CACHE: bool = False
	OBJECT_STORAGE_NEED_DATE_DIRECTORY: bool = False

	SQL_ANALYZER_ON: bool = True
	
	class Config:
		# secrets_dir = '/var/run'
		env_file = '.env'
		env_file_encoding = 'utf-8'
		case_sensitive = True


storage_settings = StorageSettings()
# logger.info(f'Storage settings[{storage_settings.dict()}].')


def ask_sql_analyzer_on() -> bool:
	return storage_settings.SQL_ANALYZER_ON

def ask_decimal_integral_digits() -> int:
	return storage_settings.DECIMAL_INTEGRAL_DIGITS


def ask_decimal_fraction_digits() -> int:
	return storage_settings.DECIMAL_FRACTION_DIGITS


def ask_disable_compiled_cache() -> bool:
	return storage_settings.DISABLE_COMPILED_CACHE


def ask_object_storage_need_date_directory() -> bool:
	return storage_settings.OBJECT_STORAGE_NEED_DATE_DIRECTORY


def ask_store_json_in_clob() -> bool:
	return storage_settings.STORE_JSON_IN_CLOB
